(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Evaluator2.sml
 *  o Description: .
 *)


structure Evaluator2 = struct

open EvaluatorsBackground

val empty_env = mk_term ("env", []) []
(* NOTE: Where is the env comming from? *)

(* EVALUATOR2 - closure conversion *)
fun Evaluator2a' steps cbv_all term env =
    let fun Eval steps term env(**) =
	    let val op_term = opid_of_term term
	    in if member op_term ["add", "subtract", "multiply", "divide", "remainder"]
	       then let val (opid, [t1, t2]) = dest_simple_term term
			val (v1, env1(**), steps1) = Eval steps  t1 env(**)
			val (v2, env2(**), steps2) = Eval steps1 t2 env(**)
		    in let val steps3 = decr_steps steps2
		       in (do_primitive_int_op opid v1 v2, empty_env(* NOTE: Why not env? *), steps3)
		       end
		       handle _ => (mk_simple_term opid [closure2term (v1, env1)(**),
							 closure2term (v2, env2)(**)],
				    env(**),
				    steps2)
		    (* NOTE: Can't we close at the leaves, i.e., for the canonical forms?
		     * Then we don't need to return the environment, do we?
		     * The environments env1 and env2 should be empty here, right? *)
		    end
	       else if op_term = "minus"
	       then let val (opid, [t1]) = dest_simple_term term
			val (v, env1(**), steps1) = Eval steps t1 env(**)
		    in if is_term "natural_number" v
		       then (mk_simple_term op_term [v], env1(**), steps1)
		       else let val steps2 = decr_steps steps1
			    in (do_primitive_minus v, env1(* TODO: Why not env? *), steps2)
			    end
			    handle _ => (mk_simple_term opid [v], env(**), steps1)
		    (* NOTE: Why not 'closure2term (v, env1)' here?
		     * Because we know that it's empty? *)
		    end
	       else if member op_term ["less", "int_eq"]
	       then let val (opid, [t1, t2, t3, t4]) = dest_simple_term term
			val (v1, env1(**), steps1) = Eval steps  t1 env(**)
			val (v2, env2(**), steps2) = Eval steps1 t2 env(**)
		    in let val steps3 = decr_steps steps2
		       in if do_primitive_cmp opid v1 v2
			  then Eval steps3 t3 env(**)
			  else Eval steps3 t4 env(**)
		       end handle _ => (mk_simple_term opid [closure2term (v1, env1)(**),
							     closure2term (v2, env2)(**),
							     t3,
							     t4],
					env(**),
					steps2)
		    end
	       else if op_term = "atom_eq"
	       then let val (operator, [([], t1),
					([], t2),
					([], t3),
					([], t4)]) = dest_term term
			val (v1, env1(**), steps1) = Eval steps  t1 env(**)
			val (v2, env2(**), steps2) = Eval steps1 t2 env(**)
		    in let val steps3 = decr_steps steps2
			   val n = (firstnat term) handle _ => 0
		       in if compare_atomn n v1 v2
			  then Eval steps3 t3 env(**)
			  else Eval steps3 t4 env(**)
		       end handle _ => (mk_term operator [([], closure2term (v1, env1)(**)),
							  ([], closure2term (v2, env2)(**)),
							  ([], t3),
							  ([], t4)],
					env(**),
					steps2)
		    end
	       else if member op_term ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
	       (* NOTE: Where are the actual "isinr", "isinl", ...?
		* Do they fall in the last else?
		* In Mark's second evaluator, there's only 'isint' and 'ispair' in here. *)
	       then let val (opid, [t1, t2, t3]) = dest_simple_term term
			val (v1, env1(**), steps1) = Eval steps t1 env(**)
		    in let val steps2 = decr_steps steps1
		       in if do_primitive_test opid v1
			  then Eval steps2 t2 env(**)
			  else Eval steps2 t3 env(**)
		       end handle _ => (mk_simple_term opid [closure2term (v1, env1)(**),
							     t2,
							     t3],
					env(**),
					steps1)
		    end
	       else if op_term = "spread"
	       then let val (operator, [([], pair), ([x, y], B)]) = dest_term term
			val (v1, env1(**), steps1) = Eval steps pair env(**)
		    in let val steps2 = decr_steps steps1
			   val (a, b) = dest_pair v1
			   val env2 = add_env_binding env x a env1(**)
			   val env3 = add_env_binding env2 y b env1(**)
			in Eval steps2 B env3(**)
			end handle _ => (mk_term operator [([], closure2term (v1, env1)(**)),
							   ([x, y], B)],
					 env(**),
					 steps1)
		    end
	       else if op_term = "decide"
	       then let val (operator, [([], d), ([x], A), ([y], B)]) = dest_term term
			val (v, env1(**), steps1) = Eval steps d env(**)
		    in let val steps2 = decr_steps steps1
		       in if is_term "inl" v
			  then let val env2 = add_env_binding env x (subtermn 1 v) env1(**)
			       in Eval steps2 A env2(**)
			       end
			  else if is_term "inr" v
			  then let val env2 = add_env_binding env y (subtermn 1 v) env1(**)
			       in Eval steps2 B env2(**)
			       end
			  else raise FAIL
		       end handle _ => (mk_term operator [([], closure2term (v, env1))(**),
							  ([x], A),
							  ([y], B)],
					env(**),
					steps1)
		    end
	       else if op_term = "apply"
	       then if steps < 0 andalso is_term "ycomb" (subtermn 1 term)
		    (* NOTE: Why do that only in negative-steps mode? *)
		    then let val steps1 = decr_steps steps
			     val f = subtermn 2 term
			 in Eval steps1 (mk_apply_term f term) env(**)
			 end handle _ => (term, env(**), steps)
		    else let val (operator, [([], function), ([], arg)]) = dest_term term
			     val (f, fenv, steps1) = Eval steps function env(**)
			 in let val steps2 = decr_steps steps1
				val (x, B) = dest_lambda f
				val env' = add_env_binding fenv x arg env(**)
			    (* NOTE: Why 'fenv' first?
			     * That's the environment in which B is evaluated. *)
			    in Eval steps2 B env'(**)
			    end handle _ => (mk_apply_term (closure2term (f, fenv)(**)) arg, env(**), steps1)
			 end
	       else if op_term = "callbyvalue"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			val (v, env1(**), steps1) = Eval steps arg env(**)
		    in let val steps2 = decr_steps steps1
		       in if is_primitive_value v
			  then let val env' = add_env_binding env x v env1(**)
			       in Eval steps2 B env'(**)
			       end
			  else raise FAIL
		       end handle _ => (mk_term operator [([], closure2term (v, env1)(**)),
							  ([x], B)],
					env(**),
					steps1)
		    end
	       else if op_term = "callbyvalueall"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			val (v1, env1(**), steps1) = Evaluator2a' steps true arg env(**)
		    in let val steps2 = decr_steps steps1
		       in if is_complete_primitive_value v1
			  then let val env' = add_env_binding env x v1 env1(**)
			       in Evaluator2a' steps2 cbv_all B env'(**)
			       end
			  else raise FAIL
		       end handle _ => (mk_term operator [([], closure2term (v1, env1)),
							  ([x], B)],
					env(**),
					steps1)
		    end
	       else if op_term = "list_ind"
	       then let val (operator, [([], L),
					([], nilcase),
					([x, xs, r], conscase)]) = dest_term term
			val (v, env1(**), steps1) = Eval steps L env(**)
		    in let val steps2 = decr_steps steps1
		       in if is_term "nil" v
			  then Eval steps2 nilcase env(**)
			  else if is_term "cons" v
			  then let val va   = subtermn 1 v
				   val vb   = subtermn 2 v
				   val env2 = add_env_binding env xs vb env1(**)
				   val env3 = add_env_binding env2 x va env1(**)
				   val w    = fresh_variable term
				   val env4 = add_env_binding env w vb env1
				   val t2 = mk_term operator [([],         mk_variable_term w),
							      ([],         nilcase),
							      ([x, xs, r], conscase)]
				   val env5 = add_env_binding env3 r t2 env4
				   (* NOTE: why not stick vb in t2 instead of a variable and create a binding:
				    * add_env_binding env3 r t2 env1.
				    * I'm doing that now. *)
				   (*val env4 = if occurs_free xs nilcase (**)
					      then let val w = maybe_new_var xs (free_vars nilcase)(**)
						       val t2 = mk_term operator [([],         mvt w),
										  ([],         nilcase),
										  ([x, xs, r], conscase)](**)
						       val recenv = add_env_binding env w vb env1(**)
						   in add_env_binding env3 r t2 recenv(**)
						   end
					      else let val t2 = mk_term operator [([],         mvt xs),
										  ([],         nilcase),
										  ([x, xs, r], conscase)](**)
						   in add_env_binding env3 r t2 env2(**)
						   end*)
				   val _ = "++\n"
				   val t = Eval steps2 conscase env5(**)
				   val _ = "--\n"
			       in t
			       end
			  else raise FAIL
		       end handle _ => (mk_term operator [([],         closure2term (v, env1)(**)),
							  ([],         nilcase),
							  ([x, xs, r], conscase)],
					env(**),
					steps1)
		    end
	       else if op_term = "ind"
	       then let val (operator, [([], i),
					([x, rd], downcase),
					([], basecase),
					([y, ru], upcase)]) = dest_term term
			val (v, qenv(**), steps1) = Eval steps i env(**)
		    in let val steps2 = decr_steps steps1
			   val n = dest_integer v
		       in if n = 0
			  then Eval steps2 basecase env(**)
			  else if n > 0
			  then let val t2 = mk_term operator [([],      mk_integer_term (n-1)),
							      ([x, rd], downcase),
							      ([],      basecase),
							      ([y, ru], upcase)]
				   val env1 = add_env_binding env y v qenv(**)
				   val env2 = add_env_binding env1 ru t2 env(**)
			       in Eval steps2 upcase env2(**)
			       end
			  else let val t2 = mk_term operator [([],      mk_integer_term (n+1)),
							      ([x, rd], downcase),
							      ([],      basecase),
							      ([y, ru], upcase)]
				   val env1 = add_env_binding env x v qenv(**)
				   val env2 = add_env_binding env1 rd t2 env(**)
			       in Eval steps2 downcase env2(**)
			       end
		       end handle _ => (mk_term operator [([],      closure2term (v, qenv)(**)),
							  ([x, rd], downcase),
							  ([],      basecase),
							  ([y, ru], upcase)],
					env(**),
					steps1)
		    end
	       else if op_term = "rec_ind"
	       then let val (operator, [([], arg), ([f, x], B)]) = dest_term term
			val function = mk_lambda_term x (mk_term operator [([], mk_variable_term x), ([f, x], B)])
		    in let val steps1 = decr_steps steps
			   val env1 = add_env_binding env x arg env(**)
			   val env2 = add_env_binding env1 f function env(**)
		       in Eval steps1 B env2(**)
		       end handle _ => (term, env, steps)
		    end
	       else if op_term = "limited_type_case"
	       then let val (v, ve(**), steps1) = Eval steps (subtermn 1 term) env(**)
		    in let val steps2 = decr_steps steps1
			   val (t2, env2(**)) = limited_type_case2(**) term v ve env(**)
		       in Eval steps2 t2 env2(**)
		       end handle _ => (limited_type_case_no_redex2(**) term v ve(**), env(**), steps1)
		    end
	       else if op_term = "TERMOF"
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (apply_conv (TagC (mk_tag_term 1)) term) env(**)
		    end handle _ => (term, env(**), steps)
	       else if false (*is_abstraction_term term*)
	       then let val steps1 = decr_steps steps
		    in Eval steps1 (unfold_ab term) env(**)
		    end handle _ => (term, env(**), steps)
	       else if op_term = "variable"
	       then let val (a, e) = lookup_binding env (dest_variable term)(**)
		    in Eval steps a e(**)
		    end handle _ => (term, env(**), steps)
	       else if member op_term ["pair", "cons"]
	       then if cbv_all
		    then let val (opid, [t1, t2]) = dest_simple_term term
			     val (v1, env1, steps1) = Eval steps  t1 env
			     val (v2, env2, steps2) = Eval steps1 t2 env
			 in (mk_simple_term opid [closure2term (v1, env1),
						  closure2term (v2, env2)],
			     env,
			     steps2)
			 end
		    else (term, env, steps)
	       else if member op_term ["inl", "inr"]
	       then if cbv_all
		    then let val (opid, [t1]) = dest_simple_term term
			     val (v1, env1, steps1) = Eval steps t1 env
			 in (mk_simple_term opid [closure2term (v1, env1)], env, steps1)
			 end
		    else (term, env, steps)
	       else (term, env, steps)
	    end
    in Eval steps term env
    end

fun Evaluator2a steps term =
    let val (term', env'(**), steps') = Evaluator2a' steps false term empty_env
    in (closure2term (term', env')(**), steps - steps')
    end


fun mk_closure_term term env = mk_simple_term "closure" [term, env]

fun mk_env_term (v, bind) env = mk_term ("env", []) [([v], bind), ([], env)]

fun empty_closure term = mk_closure_term term empty_env

fun is_closure_term term = mk_primitive_op_instance_recognizer "closure" term

fun dest_closure (TERM (("closure", []), [B_TERM ([], term), B_TERM ([], env)])) = (term, env)
  | dest_closure term = raise FAIL_WITH "dest_closure" (*(term, empty_env)*)

(*fun lookup_binding_envs [] v = raise Fail "lookup_binding_envs"
  | lookup_binding_envs (env :: envs) v =
    lookup_binding env v handle _ => lookup_binding_envs envs v*)

fun append_closure (TERM (("env", []), [])) env = env
  | append_closure (TERM (("env", []), [B_TERM ([v], bind), B_TERM ([], env)])) env' =
    mk_env_term (v, bind) (append_closure env env')
  | append_closure _ _ = raise FAIL_WITH "append_closure"

fun app_closure term bounds env =
    if is_closure_term term
    then let val (t, e) = dest_closure term
	 in app_closure t bounds (append_closure e env)
	 end
    else if is_variable_term term
    then let val v = dest_variable term
	 in if List.exists (fn x => x = v) bounds
	    then term
	    else let val (t, e) = lookup_binding env v
		 in app_closure t bounds e
		 end
		 handle _ => term
	 end
    else let val (opr, bterms) = dest_term term
	     val bterms' =
		 map (fn (vars, term) =>
			 (vars, app_closure term (vars @ bounds) env))
		     bterms
	 in mk_term opr bterms'
	 end

fun apply_closure term = app_closure term [] empty_env

(* EVALUATOR2 - closure conversion *)
fun Evaluator2b' steps cbv_all closure =
    let fun EvalTerm steps term env = Eval steps (mk_closure_term term env)
	and Eval steps closure =
	    let val (term, env) = dest_closure closure
		val op_term = opid_of_term term
	    in if op_term = "closure"
	       then let val (t, e) = dest_closure term
		    in EvalTerm steps t (append_closure e env)
		    end
	       else if member op_term ["add", "subtract", "multiply", "divide", "remainder"]
	       then let val (opid, [t1, t2]) = dest_simple_term term
			val (closure1, steps1) = EvalTerm steps  t1 env
			val (closure2, steps2) = EvalTerm steps1 t2 env
		    in let val steps3 = decr_steps steps2
			   val (v1, _) = dest_closure closure1
			   val (v2, _) = dest_closure closure2
		       in (empty_closure (do_primitive_int_op opid v1 v2), steps3)
		       end
		       handle _ =>
			      let val term = mk_simple_term
						 opid
						 [apply_closure closure1,
						  apply_closure closure2]
			      in (mk_closure_term term env, steps2)
			      end
		    end
	       else if op_term = "minus"
	       then let val (opid, [t1]) = dest_simple_term term
			val (closure1, steps1) = EvalTerm steps t1 env
		    in let val (v, _) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if is_term "natural_number" v
			  then (empty_closure (mk_simple_term op_term [v]), steps1)
			  else (empty_closure (do_primitive_minus v), steps2)
		       end
		       handle _ =>
			      let val term = mk_simple_term opid [apply_closure closure1]
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if member op_term ["less", "int_eq"]
	       then let val (opid, [t1, t2, t3, t4]) = dest_simple_term term
			val (closure1, steps1) = EvalTerm steps  t1 env
			val (closure2, steps2) = EvalTerm steps1 t2 env
		    in let val (v1, _) = dest_closure closure1
			   val (v2, _) = dest_closure closure2
			   val steps3  = decr_steps steps2
		       in if do_primitive_cmp opid v1 v2
			  then EvalTerm steps3 t3 env
			  else EvalTerm steps3 t4 env
		       end
		       handle _ =>
			      let val term = mk_simple_term
						 opid
						 [apply_closure closure1,
						  apply_closure closure2,
						  t3,
						  t4]
			      in (mk_closure_term term env, steps2)
			      end
		    end
	       else if op_term = "atom_eq"
	       then let val (opr, [([], t1),
				   ([], t2),
				   ([], t3),
				   ([], t4)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps  t1 env
			val (closure2, steps2) = EvalTerm steps1 t2 env
		    in let val steps3 = decr_steps steps2
			   val (v1, _) = dest_closure closure1
			   val (v2, _) = dest_closure closure2
			   val n = firstnat term handle _ => 0
		       in if compare_atomn n v1 v2
			  then EvalTerm steps3 t3 env
			  else EvalTerm steps3 t4 env
		       end
		       handle _ =>
			      let val term =
				      mk_simple_op_term
					  opr
					  [apply_closure closure1,
					   apply_closure closure2,
					   t3,
					   t4]
			      in (mk_closure_term term env, steps2)
			      end
		    end
	       else if member op_term ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
	       then let val (opid, [t1, t2, t3]) = dest_simple_term term
			val (closure1, steps1) = EvalTerm steps t1 env
		    in let val (v1, _) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if do_primitive_test opid v1
			  then EvalTerm steps2 t2 env
			  else EvalTerm steps2 t3 env
		       end
		       handle _ =>
			      let val term =
				      mk_simple_term
					  opid
					  [apply_closure closure1,
					   t2,
					   t3]
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "spread"
	       then let val (operator, [([], pair), ([x, y], B)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps pair env
		    in let val (v1, env1) = dest_closure closure1
			   val steps2 = decr_steps steps1
			   val (a, b) = dest_pair v1
			   val env2 = add_env_binding env  x a env1
			   val env3 = add_env_binding env2 y b env1
			in EvalTerm steps2 B env3
			end
		       handle _ =>
			      let val term =
				      mk_spread_term
					  (apply_closure closure1)
					  (x, y, B)
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "decide"
	       then let val (operator, [([], d), ([x], A), ([y], B)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps d env
		    in let val (v1, env1) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if is_term "inl" v1
			  then let val env2 = add_env_binding env x (subtermn 1 v1) env1
			       in EvalTerm steps2 A env2
			       end
			  else if is_term "inr" v1
			  then let val env2 = add_env_binding env y (subtermn 1 v1) env1
			       in EvalTerm steps2 B env2
			       end
			  else raise FAIL
		       end
		       handle _ =>
			      let val term = mk_decide_term
						 (apply_closure closure1)
						 (x, A)
						 (y, B)
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "apply"
	       then if steps < 0 andalso is_term "ycomb" (subtermn 1 term)
		    then let val steps1 = decr_steps steps
			     val f = subtermn 2 term
			 in EvalTerm steps1 (mk_apply_term f term) env
			 end
			 handle _ => (closure, steps)
		    else let val (operator, [([], function), ([], arg)]) = dest_term term
			     val (closure1, steps1) = EvalTerm steps function env
			 in let val (f, fenv) = dest_closure closure1
				val steps2 = decr_steps steps1
				val (x, B) = dest_lambda f
				val env' = add_env_binding fenv x arg env
			    in EvalTerm steps2 B env'
			    end
			    handle _ =>
				   let val term = mk_apply_term
						      (apply_closure closure1)
						      arg
				   in (mk_closure_term term env, steps1)
				   end
			 end
	       else if op_term = "callbyvalue"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps arg env
		    in let val (v1, env1) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if is_primitive_value v1
			  then let val env' = add_env_binding env x v1 env1
			       in EvalTerm steps2 B env'
			       end
			  else raise FAIL
		       end
		       handle _ =>
			      let val term = mk_callbyvalue_term
						 (apply_closure closure1)
						 (x, B)
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "callbyvalueall"
	       then let val (operator, [([], arg), ([x], B)]) = dest_term term
			val (closure1, steps1) = Evaluator2b' steps true (mk_closure_term arg env)
		    in let val (v1, env1) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if is_complete_primitive_value v1
			  then let val env' = add_env_binding env x v1 env1
			       in Evaluator2b' steps2 cbv_all (mk_closure_term B env')
			       end
			  else raise FAIL
		       end
		       handle _ =>
			      let val term = mk_callbyvalueall_term
						 (apply_closure closure1)
						 (x, B)
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "list_ind"
	       then let val (operator, [([], L),
					([], nilcase),
					([x, xs, r], conscase)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps L env
		    in let val (v1, env1) = dest_closure closure1
			   val steps2 = decr_steps steps1
		       in if is_term "nil" v1
			  then EvalTerm steps2 nilcase env
			  else if is_term "cons" v1
			  then let val va   = subtermn 1 v1
				   val vb   = subtermn 2 v1
				   val env2 = add_env_binding env xs vb env1
				   val env3 = add_env_binding env2 x va env1
				   val w    = fresh_variable term
				   val env4 = add_env_binding env w vb env1
				   val t2 = mk_term operator [([],         mk_variable_term w),
							      ([],         nilcase),
							      ([x, xs, r], conscase)]
				   val env5 = add_env_binding env3 r t2 env4
			       in EvalTerm steps2 conscase env5
			       end
			  else raise FAIL
		       end handle _ =>
				  let val term = mk_list_ind_term
						     (apply_closure closure1)
						     nilcase
						     (x, xs, r, conscase)
				  in (mk_closure_term term env, steps1)
				  end
		    end
	       else if op_term = "ind"
	       then let val (operator, [([], i),
					([x, rd], downcase),
					([], basecase),
					([y, ru], upcase)]) = dest_term term
			val (closure1, steps1) = EvalTerm steps i env
		    in let val (v, qenv) = dest_closure closure1
			   val steps2 = decr_steps steps1
			   val n = dest_integer v
		       in if n = 0
			  then EvalTerm steps2 basecase env
			  else if n > 0
			  then let val t2 = mk_term operator [([],      mk_integer_term (n-1)),
							      ([x, rd], downcase),
							      ([],      basecase),
							      ([y, ru], upcase)]
				   val env1 = add_env_binding env y v qenv
				   val env2 = add_env_binding env1 ru t2 env
			       in EvalTerm steps2 upcase env2
			       end
			  else let val t2 = mk_term operator [([],      mk_integer_term (n+1)),
							      ([x, rd], downcase),
							      ([],      basecase),
							      ([y, ru], upcase)]
				   val env1 = add_env_binding env x v qenv
				   val env2 = add_env_binding env1 rd t2 env
			       in EvalTerm steps2 downcase env2
			       end
		       end
		       handle _ =>
			      let val term = mk_ind_term
						 (apply_closure closure1)
						 (x, rd, downcase)
						 basecase
						 (y, ru, upcase)
			      in (mk_closure_term term env, steps1)
			      end
		    end
	       else if op_term = "rec_ind"
	       then let val (operator, [([], arg), ([f, x], B)]) = dest_term term
			val function = mk_lambda_term x (mk_term operator [([], mk_variable_term x), ([f, x], B)])
		    in let val steps1 = decr_steps steps
			   val env1 = add_env_binding env x arg env
			   val env2 = add_env_binding env1 f function env
		       in EvalTerm steps1 B env2
		       end handle _ => (closure, steps)
		    end
	       else if op_term = "limited_type_case"
	       then let val (closure1, steps1) = EvalTerm steps (subtermn 1 term) env
		    in let val (v, ve) = dest_closure closure1
			   val steps2 = decr_steps steps1
			   val (t2, env2) = limited_type_case2 term v ve env
		       in EvalTerm steps2 t2 env2
		       end
		       handle _ =>
			      let val term = limited_type_case_no_redex2 term (apply_closure closure1) env
			      in (mk_closure_term term  env, steps1)
			      end
		    end
	       else if op_term = "TERMOF"
	       then let val steps1 = decr_steps steps
		    in EvalTerm steps1 (apply_conv (TagC (mk_tag_term 1)) term) env
		    end
		    handle _ => (closure, steps)
	       else if false (*is_abstraction_term term*)
	       then let val steps1 = decr_steps steps
		    in EvalTerm steps1 (unfold_ab term) env
		    end
		    handle _ => (closure, steps)
	       else if op_term = "variable"
	       then let val (t, e) = lookup_binding env (dest_variable term)
		    in EvalTerm steps t e
		    end
		    handle _ => (closure, steps)
	       else if member op_term ["pair", "cons"]
	       then if cbv_all
		    then let val (opid, [t1, t2]) = dest_simple_term term
			     val (closure1, steps1) = EvalTerm steps  t1 env
			     val (closure2, steps2) = EvalTerm steps1 t2 env
			     val term = mk_simple_term opid [closure1, closure2]
			 in (empty_closure term, steps2)
			 end
		    else (closure, steps)
	       else if member op_term ["inl", "inr"]
	       then if cbv_all
		    then let val (opid, [t1]) = dest_simple_term term
			     val (closure1, steps1) = EvalTerm steps t1 env
			 in let val (v1, env1) = dest_closure closure1
				val term = mk_simple_term opid [v1]
			    in (mk_closure_term term env1, steps1)
			    end
			    handle _ =>
				   let val term = mk_simple_term
						  opid
						  [closure1]
				   in (empty_closure term, steps1)
				   end
			 end
		    else (closure, steps)
	       else (closure, steps)
	    end
    in Eval steps closure
    end

fun Evaluator2b steps term =
    let val closure = empty_closure term
	val (closure', steps') = Evaluator2b' steps false closure
    in (apply_closure closure', steps - steps')
    end

val Evaluator2 = Evaluator2b

(*
fun  Evaluator2' steps cbv_all term env =
     let fun Eval steps term env =
	      let val  op_term = opid_of_term term
	      in if member op_term ["add", "subtract", "multiply", "divide", "remainder"]
		 then let val  (opid, [t1, t2]) = dest_simple_term term
			  val  (v1, env1, steps1) = Eval steps t1 env
			  val  (v2, env2, steps2) = Eval steps1 t2 env
		      in let val  steps3 = decr_steps steps2
			 in (do_primitive_int_op opid v1 v2, env, steps3)
			 end handle _ => (mk_simple_term opid [t1, t2], env, steps2)
		      end
		 else if op_term = "minus"
		 then let val  (opid, [t1]) = dest_simple_term term
			  val  (v, env1, steps1) = Eval steps t1 env
		      in if is_term "natural_number" v
			 then (mk_simple_term op_term [v], env, steps1)
			 else let val  steps2 = decr_steps steps1
			      in (do_primitive_minus v, env, steps2)
			      end handle _ => (mk_simple_term opid [closure2term (v, env1)], env, steps1)
		      end
		 else if member op_term ["less", "int_eq"]
		 then let val  (opid, [t1, t2, t3, t4]) = dest_simple_term term
			  val  (v1, env1, steps1) = Eval steps t1 env
			  val  (v2, env2, steps2) = Eval steps1 t2 env
		      in let val  steps3 = decr_steps steps2
			 in if do_primitive_cmp opid v1 v2
			    then Eval steps3 t3 env
			    else Eval steps3 t4 env
			 end handle _ => (mk_simple_term opid [closure2term (v1, env1), closure2term (v2, env2), t3, t4], env, steps2)
		      end
		 else if op_term = "atom_eq"
		 then let val  (operator, [([], t1), ([], t2), ([], t3), ([], t4)]) = dest_term term
			  val  (v1, env1, steps1) = Eval steps t1 env
			  val  (v2, env2, steps2) = Eval steps1 t2 env
		      in let val  steps3 = decr_steps steps2
			     val  n = (firstnat term) handle _ => 0
			 in if compare_atomn n v1 v2
			    then Eval steps3 t3 env
			    else Eval steps3 t4 env
			 end handle _ => (mk_term operator [([], closure2term (v1, env1)), ([], closure2term (v2, env2)), ([], t3), ([], t4)], env, steps2)
		      end
		 else if member op_term ["isinr", "isinl", "ispair", "isint", "islambda", "isatom2"]
		 then let val  (opid, [t1, t2, t3]) = dest_simple_term term
			  val  (v1, env1, steps1) = Eval steps t1 env
		      in let val  steps2 = decr_steps steps1
			 in if do_primitive_test opid v1
			    then Eval steps2 t2 env
			    else Eval steps2 t3 env
			 end handle _ => (mk_simple_term opid [closure2term (v1, env1), t2, t3], env, steps1)
		      end
		 else if op_term = "spread"
		 then let val  (operator, [([], pair), ([x, y], B)]) = dest_term term
			  val  (v1, env1, steps1) = Eval steps pair env
		      in let val  steps2 = decr_steps steps1
			     val  (a, b) = dest_pair v1
			 in let val  env' = add_env_binding env x a env1
				val  env'' = add_env_binding env' y b env1
			    in Eval steps2 B env''
			    end
			 end handle _ => (mk_term operator [([], closure2term (v1, env1)), ([x, y], B)], env, steps1)
		      end
		 else if op_term = "decide"
		 then let val  (operator, [([], d), ([x], A), ([y], B)]) = dest_term term
			  val  (v, env1, steps1) = Eval steps d env
		      in let val  steps2 = decr_steps steps1
			 in if is_term "inl" v
			    then let val  env' = add_env_binding env x (subtermn 1 v) env1
				 in Eval steps2 A env'
				 end
			    else if is_term "inr" v
			    then let val  env' = add_env_binding env y (subtermn 1 v) env1
				 in Eval steps2 B env'
				 end
			    else raise FAIL
			 end handle _ => (mk_term operator [([], closure2term (v, env1)), ([x], A), ([y], B)], env, steps1)
		      end
		 else if op_term = "apply"
		 then if steps < 0 andalso is_term "ycomb" (subtermn 1 term)
		      then let val  steps1 = decr_steps steps
			       val  f = subtermn 2 term
			   in Eval steps1 (mk_apply_term f term) env
			   end handle _ => (term, env, steps)
		      else let val  (operator, [([], function), ([], arg)]) = dest_term term
			       val  (f, env1, steps1) = Eval steps function env
			   in let val  steps2 = decr_steps steps1
				  val  (x, B) = dest_lambda f
			      in let val  env1' = add_env_binding env1 x arg env
				 in Eval steps2 B env1'
				 end
			      end handle _ => (mk_apply_term f arg, env, steps1)
			   end
		 else if op_term = "callbyvalue"
		 then let val  (operator, [([], arg), ([x], B)]) = dest_term term
			  val  (v, env1, steps1) = Eval steps arg env
		      in let val  steps2 = decr_steps steps1
			 in if is_primitive_value v
			    then let val  env' = add_env_binding env x v env1
				 in Eval steps2 B env'
				 end
			    else raise FAIL
			 end handle _ => (mk_term operator [([], closure2term (v, env1)), ([x], B)], env, steps1)
		      end
		 else if op_term = "callbyvalueall"
		 then let val  (operator, [([], arg), ([x], B)]) = dest_term term
			  val  (v1, env1, steps1) = Evaluator2' steps true arg env
		      in let val  steps2 = decr_steps steps1
			 in if is_complete_primitive_value v1
			    then let val  env' = add_env_binding env x v1 env1
				 in Evaluator2' steps2 cbv_all B env'
				 end
			    else raise FAIL
			 end handle _ => (mk_term operator [([], closure2term (v1, env1)), ([x], B)], env, steps1)
		      end
		 else if op_term = "list_ind"
		 then let val  (operator, [([], L), ([], nilcase), ([x, xs, r], conscase)]) = dest_term term
			  val  (v, env1, steps1) = Eval steps L env
		      in let val  steps2 = decr_steps steps1
			 in if is_term "nil" v
			    then Eval steps2 nilcase env
			    else if is_term "cons" v
			    then let val  va = subtermn 1 v
				     val  vb = subtermn 2 v
				     val  w = fresh_variable term
				     val  envr = add_env_binding env w vb env1
				     val  t2 = mk_term operator [([], mk_variable_term w), ([], nilcase), ([x, xs, r], conscase)]
				 in let val  env' = add_env_binding env x va env1
					val  env'' = add_env_binding env' xs vb env1
					val  env''' = add_env_binding env'' r t2 envr
				    in Eval steps2 conscase env'''
				    end
				 end
			    else raise FAIL
			 end handle _ => (mk_term operator [([], closure2term (v, env1)), ([], nilcase), ([x, xs, r], conscase)], env, steps1)
		      end
		 else if op_term = "ind"
		 then let val  (operator, [([], i), ([x, rd], downcase), ([], basecase), ([y, ru], upcase)]) = dest_term term
			  val  (v, env1, steps1) = Eval steps i env
		      in let val  steps2 = decr_steps steps1
			     val  n = dest_integer v
			 in if n = 0
			    then Eval steps2 basecase env
			    else if n > 0
			    then let val  t2 = mk_term operator [([], mk_integer_term (n - 1)), ([x, rd], downcase), ([], basecase), ([y, ru], upcase)]
				 in let val  env' = add_env_binding env y v env1
					val  env'' = add_env_binding env' ru t2 env
				    in Eval steps2 upcase env''
				    end
				 end
			    else let val  t2 = mk_term operator [([], mk_integer_term (n + 1)), ([x, rd], downcase), ([], basecase), ([y, ru], upcase)]
				 in let val  env' = add_env_binding env x v env1
					val  env'' = add_env_binding env' rd t2 env
				    in Eval steps2 downcase env''
				    end
				 end
			 end handle _ => (mk_term operator [([], closure2term (v, env1)), ([x, rd], downcase), ([], basecase), ([y, ru], upcase)], env, steps1)
		      end
		 else if op_term = "rec_ind"
		 then let val  (operator, [([], arg), ([f, x], B)]) = dest_term term
			  val  function = mk_lambda_term x (mk_term operator [([], mk_variable_term x), ([f, x], B)])
		      in let val  steps1 = decr_steps steps
			 in let val  env' = add_env_binding env x arg env
				val  env'' = add_env_binding env' f function env
			    in Eval steps1 B env''
			    end
			 end handle _ => (term, env, steps)
		      end
		 else if op_term = "limited_type_case"
		 then let val  (v, env1, steps1) = Eval steps (subtermn 1 term) env
		      in let val  steps2 = decr_steps steps1
			 in Eval steps2 (limited_type_case1 term v) env
			 end handle _ => (limited_type_case_no_redex1 term v, env, steps1)
		      end
		 else if op_term = "TERMOF"
		 then let val  steps1 = decr_steps steps
		      in Eval steps1 (apply_conv (TagC (mk_tag_term 1)) term) env
		      end handle _ => (term, env, steps)
		 else if false
		 then let val  steps1 = decr_steps steps
		      in Eval steps1 (unfold_ab term) env
		      end handle _ => (term, env, steps)
		 else if op_term = "variable"
		 then let val  (a, e) = lookup_binding env (dest_variable term)
		      in Eval steps a e
		      end handle _ => (term, env, steps)
		 else if member op_term ["pair", "cons"]
		 then if cbv_all
		      then let val  (opid, [t1, t2]) = dest_simple_term term
			       val  (v1, env1, steps1) = Eval steps t1 env
			       val  (v2, env2, steps2) = Eval steps1 t2 env
			   in (mk_simple_term opid [v1, v2], env, steps2)
			   end
		      else (term, env, steps)
		 else if member op_term ["inl", "inr"]
		 then if cbv_all
		      then let val  (opid, [t1]) = dest_simple_term term
			       val  (v1, env1, steps1) = Eval steps t1 env
			   in (mk_simple_term opid [v1], env, steps1)
			   end
		      else (term, env, steps)
		 else (term, env, steps)
	      end
     in Eval steps term env
     end
fun  Evaluator2 steps term =
     let val  (term', env1, steps') = Evaluator2' steps false term empty_env
     in (closure2term (term', env1), steps - steps')
     end
*)

end
