(* Copyright 2011 Cornell University
 *
 *
 * This file is part of EventML - a tool aiming at specifying
 * distributed protocols in an ML like language.  It is an interface
 * to the logic of events and is compiled into Nuprl.  It is written
 * by the NUPRL group of Cornell University, Ithaca, NY.
 *
 * EventML is a free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * EventML is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with EventML.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  o Authors:     Vincent Rahli
 *  o Affiliation: Cornell University, NUPRL group
 *  o Date:        20 May 2011
 *  o File name:   Interface.sml
 *  o Description: EventML interface.
 *)


structure Interface :> INTERFACE = struct

structure P  = Parser
structure A  = Ast
structure E  = Enum
structure N  = ToNuprl
structure D  = Deps
structure G  = Gen
structure EN = Env
structure GL = GenLib
structure NT = NuprlTerms
structure EH = LibBase
structure EV = Evaluators


(* ------ ARGUMENTS ------ *)

val default_el_output = "/tmp/eventml-output.el"
val default_output    = NONE
val default_lib       = NONE
val default_input     = "test.esh"
val default_time      = LargeInt.fromInt 1000 (* 1sec by default *)
val default_sub       = false
val default_sanity    = false
val default_nuprl     = false
val default_obid      = ""
val default_ascii     = false
val default_tcheck    = false
val default_parse     = false
val default_split     = false
val default_prt       = false
val default_eval      = NONE
val default_alldef    = NONE
val default_test      = NONE
val default_session   = false

val initArgs =
    {input   = default_input,
     output  = default_output,
     lib     = default_lib,
     time    = default_time,
     sub     = default_sub,
     sanity  = default_sanity,
     nuprl   = default_nuprl,
     obid    = default_obid,
     ascii   = default_ascii,
     tcheck  = default_tcheck,
     parse   = default_parse,
     split   = default_split,
     prt     = default_prt,
     eval    = default_eval,
     alldef  = default_alldef,
     test    = default_test,
     session = default_session}

fun getElOutput (SOME output) = output
  | getElOutput NONE = default_el_output

fun updInput   {input = _, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} input   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updOutput  {input, output = _, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} output  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updLib     {input, output, lib = _, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} lib     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updTime    {input, output, lib, time = _, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} time    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updSub     {input, output, lib, time, sub = _, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} sub     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updSanity  {input, output, lib, time, sub, sanity = _, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} sanity  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updNuprl   {input, output, lib, time, sub, sanity, nuprl = _, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} nuprl   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updObid    {input, output, lib, time, sub, sanity, nuprl, obid = _, ascii, tcheck, parse, split, prt, eval, alldef, test, session} obid    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updAscii   {input, output, lib, time, sub, sanity, nuprl, obid, ascii = _, tcheck, parse, split, prt, eval, alldef, test, session} ascii   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updTcheck  {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck = _, parse, split, prt, eval, alldef, test, session} tcheck  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updParse   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse = _, split, prt, eval, alldef, test, session} parse   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updSplit   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split = _, prt, eval, alldef, test, session} split   = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updPrint   {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt = _, eval, alldef, test, session} prt     = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updEval    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval = _, alldef, test, session} eval    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updAlldef  {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef = _, test, session} alldef  = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updTest    {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test = _, session} test    = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}
fun updSession {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session = _} session = {input = input, output = output, lib = lib, time = time, sub = sub, sanity = sanity, nuprl = nuprl, obid = obid, ascii = ascii, tcheck = tcheck, parse = parse, split = split, prt = prt, eval = eval, alldef = alldef, test = test, session = session}

datatype arg = I    of string        (* input     *)
	     | O    of string        (* ouput     *)
	     | L    of string        (* library   *)
	     | T    of LargeInt.int  (* timelimit *)
	     | OBID of string        (* object_id to use when generating the nuprl code *)
	     | EVAL of string        (* term to evaluate *)
	     | DEF  of string        (* the alldef file  *)
	     | TEST of int
	     | SESSION
	     | TONUPRL
	     | SUBTYPING
	     | SANITY
	     | FROMASCII
	     | TYPECHECK
	     | PARSE
	     | SPLIT
	     | PRINT

fun format args =
    let fun gen [] r = r
	  | gen (SUBTYPING      :: list) r = gen list (updSub     r true)
	  | gen (SANITY         :: list) r = gen list (updSanity  r true)
	  | gen (FROMASCII      :: list) r = gen list (updAscii   r true)
	  | gen (TYPECHECK      :: list) r = gen list (updTcheck  r true)
	  | gen (PARSE          :: list) r = gen list (updParse   r true)
	  | gen (SPLIT          :: list) r = gen list (updSplit   r true)
	  | gen (PRINT          :: list) r = gen list (updPrint   r true)
	  | gen (TONUPRL        :: list) r = gen list (updNuprl   r true)
	  | gen (SESSION        :: list) r = gen list (updSession r true)
	  | gen ((OBID obid)    :: list) r = gen list (updObid    r obid)
	  | gen ((T    time)    :: list) r = gen list (updTime    r time)
	  | gen ((I    file)    :: list) r = gen list (updInput   r file)
	  | gen ((TEST n)       :: list) r = gen list (updTest    r (SOME n))
	  | gen ((EVAL str)     :: list) r = gen list (updEval    r (SOME str))
	  | gen ((O    file)    :: list) r = gen list (updOutput  r (SOME file))
	  | gen ((L    file)    :: list) r = gen list (updLib     r (SOME file))
	  | gen ((DEF  file)    :: list) r = gen list (updAlldef  r (SOME file))
    in gen args initArgs
    end


(* ------ INTERFACE FUNCTIONS ------ *)

fun parseEML input =
    let val _    = D.reset ()
	val term = P.parse [input]
	val st1  = A.export term
	val st2  = A.toString term
	val _    = print (st1 ^ "\n")
	val _    = print (st2 ^ "\n")
    in ()
    end

fun getLastAbstraction [] = (NONE, [])
  | getLastAbstraction lst =
    let val rev    = List.rev lst
	val last   = List.hd rev
	val firsts = List.rev (List.tl rev)
    in if NT.is_nuprl_iabstraction_term last
       then (SOME (NT.dest_iabstraction last), [])
       else let val (termop, terms) = getLastAbstraction firsts
	    in (termop, terms @ [last])
	    end
    end

fun keepAbsWf [] = []
  | keepAbsWf (term :: terms) =
    if NT.is_nuprl_iabstraction_term term
       orelse
       NT.is_nuprl_iwftheorem_term term
    then term :: keepAbsWf terms
    else keepAbsWf terms

fun pairAbsWf [] = []
  | pairAbsWf [x] = raise EH.Impossible "pairAbsWf"
  | pairAbsWf (abs :: wf :: terms) =
    if NT.is_nuprl_iabstraction_term abs
       andalso
       NT.is_nuprl_iwftheorem_term wf
    then (abs, wf) :: pairAbsWf terms
    else raise EH.Impossible "pairAbsWf"

fun splitLastAbs terms =
    let val terms' = keepAbsWf terms
	val pairs  = pairAbsWf terms'
	val rev    = List.rev pairs
    in if List.length rev > 0
       then let val (abs, wf) = List.hd rev
		val firsts = List.rev (List.tl rev)
	    in (SOME (NT.dest_iabstraction abs, wf), firsts)
	    end
       else (NONE, [])
    end

fun evaluate_gen ev input str libop obid sub prt =
    let val _        = D.reset ()
	val tmp      = OS.FileSys.tmpName ()
	val stout    = TextIO.openOut tmp
	val _        = TextIO.output (stout, "let it = " ^ str ^ "\n")
	val _        = TextIO.closeOut stout
    in let val (term, env, b) = E.slice [input, tmp] NONE libop 1 prt sub true
       in if b
	  then let val s_term   = A.simplify term
		   val n_terms1 = N.toNuprl tmp obid s_term env sub
		   val _        = OS.FileSys.remove tmp
	       in case splitLastAbs n_terms1 of
		      (SOME ((t1, t2, t3), wf), terms) =>
		      let val t3' = ev (NT.mk_nuprl_evalall_term t3) terms
			  val abs = NT.mk_nuprl_iabstraction_term t2 t3'
			  val rhs = #3 (NT.dest_iabstraction abs)
			  val thm = NT.dest_iwftheorem wf
			  val sta = NT.nuprl2eml_abs "it" abs
			  val stw = NT.nuprl2eml_wf  "it" wf
			  val _   = if sta = "`???`"
				    then print ("evaluation failed.\n")
				    else print (sta ^ " : " ^ stw ^ "\n")
		      (*app (fn t => print (NT.nuprl2eml t ^ "\n")) [abs, wf]*)
		      in ()
		      end
		    | _ => ()
	       end
	  else print "Untypable.\n"
       end handle e => (OS.FileSys.remove tmp; raise e)
    end

fun evaluate input str libop alldefop obid sub =
    evaluate_gen (fn t => fn ts => #1 (EV.run_ev1 true alldefop ts (~1) t))
		 input str libop obid sub (true, true)

fun evaluate_map input str libop obid sub =
    evaluate_gen (fn t => fn ts => #1 (EV.run_ev1_map ts (~1) t))
		 input str libop obid sub (true, false)

fun run_session strm libop input obid sub =
    case TextIO.inputLine strm of
	NONE => run_session strm libop input obid sub
      | SOME line =>
	if line = "quit\n"
	then let val _ = print "\nend of session\n"
	     in EV.end_session ()
	     end
	else let val _ = evaluate_map input line libop obid sub
		 val _ = print "EventML# "
	     in run_session strm libop input obid sub
	     end

fun start_session input libop alldefop obid sub =
    let val _    = print "Loading library...\n"
	val _    = EV.start_session false alldefop
	val strm = TextIO.stdIn
	val _    = print "EventML# "
    in run_session strm libop input obid sub
    end

val eval_test = EV.test

val typecheck = E.typecheck

fun slice input outputop libop timelimit sub =
    let val (prog, env, b) = E.slice [input] outputop libop timelimit (true, true) sub false
    in ()
    end

fun toNuprl input outputop libop obid timelimit sub prt =
    let val (term, env, b) = E.slice [input] NONE libop timelimit (true, true) sub false
    in if b
       then let val s_term  = A.simplify term
		val n_terms = N.toNuprl input obid s_term env sub
		val st      =
		    if prt
		    then app (fn n_term => print (NT.toStringTerm n_term ^ "\n")) n_terms
		    else ()
		val _       =
		    case outputop of
			NONE => ()
		      | SOME file =>
			let val file' = file ^ ".tmp"
			    val stout = TextIO.openOut file'
			    val _     = map (fn n_term => TextIO.output (stout, NT.toStringTerm n_term ^ "\004\n")) n_terms
			    val _     = TextIO.closeOut stout
			    val _     = OS.FileSys.rename {old = file', new = file}
			in () end
			handle IO.Io {name, function, cause} =>
			       (raise EH.Impossible ("cannot open or close one of the output file (function:" ^ function ^ ",name:" ^ name ^ ")\n"))
	    in ()
	    end
       else print "We won't convert the program into a Nuprl term because it contains errors.\n"
    end

fun parseNuprl input outputop prt split =
    GL.exportTerms input outputop prt split

fun sanitizer input outputop libop timelimit sub =
    let val (prog, env, b) = E.slice [input] outputop libop timelimit (true, true) sub true
    in ()
    end

val testing = eval_test

fun run args =
    let val {input, output, lib, time, sub, sanity, nuprl, obid, ascii, tcheck, parse, split, prt, eval, alldef, test, session} = format args
    in if Option.isSome test
       then testing (Option.valOf test)
       else if parse
       then parseEML input
       else if tcheck
       then typecheck input sub
       else if sanity
       then sanitizer input output lib time sub
       else if nuprl
       then toNuprl input output lib obid time sub prt
       else if ascii
       then parseNuprl input output prt split
       else if Option.isSome eval
       then evaluate input (Option.valOf eval) lib alldef obid sub
       else if session
       then start_session input lib alldef obid sub
       else slice input output lib time sub
    end

end
